/*
								+----------------------------------+
								|                                  |
								| ***  Pashik virtual machine  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2002   |
								|                                  |
								|            Lexer.cpp             |
								|                                  |
								+----------------------------------+
*/

#include "stdafx.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include "TextParser.h"

/*
 *			=== Lexical analyzer ===
 */

enum { MaxLen = 256 };
// max. dlka sla ve zancch

#define PI ((float)3.141592653589793238462643383279502884)
// pi

CLexicalAnalyzer::CLexicalAnalyzer()
{
    m_p_stream = NULL;
	m_s_last_error[0] = 0;
}

CLexicalAnalyzer::~CLexicalAnalyzer()
{
    Close_Stream();
}

void CLexicalAnalyzer::Error(char *p_str, int n_line)
{
	char p_line[16];

	sprintf(p_line, " [%d]", n_line);
	strncpy(m_s_last_error, p_str, 255);
	if(strlen(m_s_last_error) + strlen(p_line) <= 255)
		strcat(m_s_last_error, p_line);
}

int CLexicalAnalyzer::Get_LastError(char *p_buffer, int n_buffer_length)
{
	if(!m_s_last_error[0])
		return false;
	strncpy(p_buffer, m_s_last_error, n_buffer_length);
	// pokud m string s chybou nenulovou dlku, zkopruje ho

	return true;
}

// soubor ulo do bufferu,
// vrt false pokud je NULL
// / przdn / nevejde se do ram,
// jinak vrac true
int CLexicalAnalyzer::Open_Stream(FILE *p_stream_fp)
{
	if(p_stream_fp == NULL)
		return false;
	// oteve soubor

	m_n_stream_len = 0;
	while(fgetc(p_stream_fp) != EOF)
		m_n_stream_len ++;
	fseek(p_stream_fp, 0, SEEK_SET);
	// zjist dlku streamu

	if(m_n_stream_len == 0)
		return false;
	// mus mt dlku

	if((m_p_stream = new char[m_n_stream_len + 1]) == NULL)
		return false;
	// naalokuje buffer

	fread(m_p_stream, m_n_stream_len, 1, p_stream_fp);
	// pete soubor

	m_p_stream[m_n_stream_len] = 0;
	// debug

	m_n_stream_pos = 0;
	// nastav pozici, chyby na nulu

	return true;
}

int CLexicalAnalyzer::Open_Stream(char *p_stream_str, int n_stream_len)
{
	if(p_stream_str == NULL)
		return false;
	// check

	m_n_stream_len = m_n_stream_len;
	if(m_n_stream_len == 0)
		return false;
	// mus mt dlku

	if((m_p_stream = new char[m_n_stream_len + 1]) == NULL)
		return false;
	// naalokuje buffer

	memcpy(m_p_stream, p_stream_str, m_n_stream_len);
	// zkopruje

	m_p_stream[m_n_stream_len] = 0;
	// debug

	m_n_stream_pos = 0;
	// nastav pozici, chyby na nulu

	return true;
}

// vrt true, pokud je analza na konci
int CLexicalAnalyzer::b_End_of_Stream()
{
	int n_last_pos;
	char c;

	n_last_pos = m_n_stream_pos;

	do {
		if((c = GetChar()) == 0) {
			m_n_stream_pos = n_last_pos;
			return true;
		}
	} while(c == ' ' || c == '\n' || c == '\r' || c == '\t');
	// pesko whitespace
	
	m_n_stream_pos = n_last_pos;

	return false;
}

// vrt true, pokud je oteveno
int CLexicalAnalyzer::b_Stream()
{
	return m_n_stream_len != 0;
}

// uvoln buffer
int CLexicalAnalyzer::Close_Stream()
{
	if(m_p_stream == NULL)
		return false;

	delete[] m_p_stream;
	m_p_stream = NULL;

	return true;
}

// vrt znak
void CLexicalAnalyzer::PushChar()
{
	if(m_n_stream_pos > 0)
		m_n_stream_pos --;
}

// pete znak (vynechv komente)
// vrac 0 pokud nen oteveno, nebo je konec
int CLexicalAnalyzer::GetChar()
{
	if(m_p_stream == NULL)
		return false;

	if(m_n_stream_pos == m_n_stream_len)
		return false;

	if(m_p_stream[m_n_stream_pos ++] != '/')
		return m_p_stream[m_n_stream_pos - 1];
	// pokud nen lomeno, vrt to (jinak by to mohl bt koment)

	if(m_p_stream[m_n_stream_pos] == '/') {
		while(m_p_stream[m_n_stream_pos] != '\n' && m_n_stream_pos != m_n_stream_len)
			m_n_stream_pos ++;
		if(m_n_stream_pos >= m_n_stream_len)
		    return false;

		return m_p_stream[m_n_stream_pos ++];
	}
	// pesko jednodkov koment //

	if(m_p_stream[m_n_stream_pos] == '*') {
        m_n_stream_pos ++;
        while(m_n_stream_pos < m_n_stream_len - 1 &&
           (m_p_stream[m_n_stream_pos] != '*' ||
           m_p_stream[m_n_stream_pos + 1] != '/'))
            m_n_stream_pos ++;
	    m_n_stream_pos += 2;
        if(m_n_stream_pos > m_n_stream_len) {
            Error("Comment into the end of file !", Cur_Line());
            return false;
        }

        return m_p_stream[m_n_stream_pos ++];
    }
    // pesko koment /* .. */

	return m_p_stream[m_n_stream_pos - 1];
}

// pete token ze streamu / vrt _Unknown_ token
Token CLexicalAnalyzer::Get_Token()
{
	int n_prev_stream_pos;
	int b_frac, b_hex;
	char text[MaxLen];
	int n_string_len;
    Token token;
	char c, x;

	token.id = _Unknown_;
	token.s_value = NULL;
	token.n_value = 0;
	token.f_value = 0;
	token.n_line = Cur_Line();
	// nastav token

	do {
		if((c = GetChar()) == 0)
		    return token;
	} while(c == ' ' || c == '\n' || c == '\r' || c == '\t');
	// pesko whitespace
	
	token.n_line = Cur_Line();
	// pokud by byl mezi posledenm a souasnm
	// identifiktorem vcedkov koment

	if(c == '.' || (c >= '0' && c <= '9')) {
		memset(text, 0, MaxLen);
		b_frac = 0;

		b_hex = 1;
		x = tolower(GetChar());
		if(!(c == '0' && x == 'x')) {
			PushChar();
			b_hex = 0;
		}
		// zjist zda je hex, kdytak pesko "0x" (v c pak bude '0')

        do {
			if(strlen(text) + 1 == MaxLen) {
				Error("Too long number found !", Cur_Line());
				return token;
			}
			//
			text[strlen(text)] = c;
			c = GetChar();
			if(c == '.')
				b_frac ++;
		} while(c != 0 && (c == '.' || (isdigit(c) || b_hex && (tolower(c) >= 'a' && tolower(c) <= 'f'))));
		if(c != 0)
			PushChar();
		// pete blok, pedstavujc slo

		if(b_hex) {
			if(b_frac)
				return token;
			// hex & desetinn nejde

			token.id = IValue;
			sscanf(text, "%x", &token.n_value);
			// pete hodnotu

			return token;
		}

		if(b_frac) {
			token.id = FValue;
			token.f_value = (float)atof(text);
			// pete hodnotu

			return token;
		}

		token.id = IValue;
		token.n_value = atol(text);
		// long - proti peteen

		return token;
	}
	// slo v hex / dec

	if(c == '\"') {
		n_prev_stream_pos = m_n_stream_pos;
		// ulo pozici ve streamu

		x = c;
		// prvn znak

		n_string_len = 1;
		// 1 pro ukonovac nulu

		while(1) {
			if(!(c = GetChar())) {
				Error("Unexpected end of file found", Cur_Line());
				return token;
			}

			if(c == '\\') {
				if(!(c = GetChar())) {
					Error("Unexpected end of file found", Cur_Line());
					return token;
				}
				switch(c) {
				case 'a':
				case 'b':
				case 'f':
				case 'n':
				case 'r':
				case 't':
				case 'v':
				case '\'':
				case '0':
				case '\"':
				case '\\':
					n_string_len ++;
					break;
				default:
					Error("Unknown escape sequence (skipping)", Cur_Line());
				}

				continue;
			} else if(c == '\"') {
				break;
			} else if(c == '\n') {
				Error("Newline in constant", Cur_Line());
				return token;
			}

			n_string_len ++;
		}
		// ur dlku stringu

		m_n_stream_pos = n_prev_stream_pos;
		// vrt pozici ve streamu

		if((token.s_value = new char[n_string_len]) == NULL) {
			Error("String is too long to fit in memory", Cur_Line());
			return token;
		}
		memset(token.s_value, 0, n_string_len);
		// vytvo przdn string

		c = x;
		while(1) {
			if(!(c = GetChar())) {
				Error("Unexpected end of file found", Cur_Line());
				return token;
			}

			if(c == '\\') {
				if(!(c = GetChar())) {
					Error("Unexpected end of file found", Cur_Line());
					return token;
				}
				switch(c) {
				case 'a':
					token.s_value[strlen(token.s_value)] = '\a';
					break;
				case 'b':
					token.s_value[strlen(token.s_value)] = '\b';
					break;
				case 'f':
					token.s_value[strlen(token.s_value)] = '\f';
					break;
				case 'n':
					token.s_value[strlen(token.s_value)] = '\n';
					break;
				case 'r':
					token.s_value[strlen(token.s_value)] = '\r';
					break;
				case 't':
					token.s_value[strlen(token.s_value)] = '\t';
					break;
				case 'v':
					token.s_value[strlen(token.s_value)] = '\v';
					break;
				case '\'':
					token.s_value[strlen(token.s_value)] = '\'';
					break;
				case '0':
					token.s_value[strlen(token.s_value)] = '\0';
					break;
				case '\"':
					token.s_value[strlen(token.s_value)] = '\"';
					break;
				case '\\':
					token.s_value[strlen(token.s_value)] = '\\';
					break;
				default:
					Error("Unknown escape sequence (skipping)", Cur_Line());
				}

				continue;
			} else if(c == '\"') {
				token.id = SValue;
				return token;
			} else if(c == '\n') {
				Error("Newline in constant", Cur_Line());
				return token;
			}

			token.s_value[strlen(token.s_value)] = c;
		}

		Error("String parsing routine error !", Cur_Line());
		return token;
	}
	// pete string

	if(c == '\'') {
		if(!(c = GetChar())) {
			Error("Unexpected end of file found", Cur_Line());
			return token;
		}
		if(c == '\\') {
			if(!(c = GetChar())) {
				Error("Unexpected end of file found", Cur_Line());
				return token;
			}
			switch(c) {
			case 'a':
				token.n_value = '\a';
				break;
			case 'b':
				token.n_value = '\b';
				break;
			case 'f':
				token.n_value = '\f';
				break;
			case 'n':
				token.n_value = '\n';
				break;
			case 'r':
				token.n_value = '\r';
				break;
			case 't':
				token.n_value = '\t';
				break;
			case 'v':
				token.n_value = '\v';
				break;
			case '\'':
				token.n_value = '\'';
				break;
			case '\"':
				token.n_value = '\"';
				break;
			case '0':
				token.n_value = '\0';
				break;
			case '\\':
				token.n_value = '\\';
				break;
			default:
				Error("Unknown escape sequence (skipping)", Cur_Line());
				token.n_value = 0;
			}
		} else if(c == '\n') {
			Error("Newline in constant", Cur_Line());
			return token;
		} else
			token.n_value = c;
		// pete character

		if(!(c = GetChar())) {
			Error("Unexpected end of file found", Cur_Line());
			return token;
		}
		if(c != '\'') {
			Error("Expected quotation mark not found", Cur_Line());
			return token;
		}
		// otestuje sprvn ukonen

		token.id = CValue;
		return token;
	}
	// pete znak (char)

	if(isalpha(c) || c == '_') {
		n_prev_stream_pos = m_n_stream_pos;
		// ulo pozici ve streamu

		x = c;
		// prvn znak

		n_string_len = 1;
		do {
			n_string_len ++;
			c = GetChar();
		} while(c != 0 && (isalpha(c) || isdigit(c) || c == '_' || c == '.' || c == '[' || c == ']'));
		if(c != 0)
			PushChar();
		// zjist velikost identifiktoru

		m_n_stream_pos = n_prev_stream_pos;
		// vrt pozici ve streamu

		if((token.s_value = new char[n_string_len]) == NULL) {
			Error("String is too long to fit in memory", Cur_Line());
			return token;
		}
		memset(token.s_value, 0, n_string_len);
		// vytvo przdn string

		c = x;
		do {
			token.s_value[strlen(token.s_value)] = c;
			c = GetChar();
		} while(c != 0 && (isalpha(c) || isdigit(c) || c == '_' || c == '.' || c == '[' || c == ']'));
		if(c != 0)
			PushChar();

		if(!strcmp(token.s_value, "_e")) {
			token.id = FValue;
			token.f_value = (float)exp(1);
		} else if(!strcmp(token.s_value, "Pi")) {
			token.id = FValue;
			token.f_value = PI;
		}
		// zkus rozpoznat token

		if(token.id != _Unknown_) {
			delete[] token.s_value;
			token.s_value = NULL;
			//
			return token;
		}
		// pokud ho nalo, vrt ho to ..

		token.id = Ident;
		// jinak to bude identifiktor

		return token;
	}
	// identifiktor / keyword

	switch(c) {
    case ',':
		token.id = Colon;
		return token;
	case '{':
		token.id = Begin;
		return token;
	case '}':
		token.id = End;
		return token;
	case ';':
		token.id = EndStmt;
		return token;
	case '=':
		token.id = Assign;
		return token;
	default:
		Error("Unknown token found", Cur_Line());
	}

	return token;
}

// najde dek
int CLexicalAnalyzer::Cur_Line()
{
	int i, n_line;

	for(i = 0, n_line = 1; i < m_n_stream_pos; i ++) {
		if(m_p_stream[i] == '\n')
			n_line ++;
	}

	return n_line;
}

// nastav m_p_stream na zatek
void CLexicalAnalyzer::Reset_Stream()
{
	m_n_stream_pos = 0;
}

/*
 *			=== Parser ===
 */

CParser::CParser()
{
	m_p_token = NULL;
	m_n_token_num = 0;
}

CParser::~CParser()
{
	Free();
}

// pete data v souboru a vytvo
// si list token, kter zkontroluje
// na chyby v posloupnosti
int CParser::Parse(FILE *p_file)
{
	int n_result;

	Free();
	m_analyzer.Open_Stream(p_file);
	n_result = Get_Tokens();
	m_analyzer.Close_Stream();
	
	return n_result;
}

int CParser::Parse(char *p_stream, int n_stream_len)
{
	int n_result;

	Free();
	m_analyzer.Open_Stream(p_stream, n_stream_len);
	n_result = Get_Tokens();
	m_analyzer.Close_Stream();

	return n_result;
}

#define __b_IsValue(x) ((x) == CValue || \
						(x) == IValue || \
						(x) == FValue || \
						(x) == SValue)
// vrt true, pokud je to hodnota

// vytvo pole token a kontroluje jeho syntaktickou sprvnost
int CParser::Get_Tokens()
{
	Token t_token;
	int i;

	m_analyzer.Reset_Stream();
	for(m_n_token_num = 0; !m_analyzer.b_End_of_Stream();) {
		m_n_token_num ++;
		t_token = m_analyzer.Get_Token();
		//
		if(t_token.id == _Unknown_) {
			m_n_token_num = 0;
			return Parser_Error;
		}
	}
	// spot tokeny

	if((m_p_token = new Token[m_n_token_num]) == NULL)
		return Parser_Memory;
	// naalokuje pole

	m_analyzer.Reset_Stream();
	for(m_n_token_num = 0; !m_analyzer.b_End_of_Stream();) {
		m_p_token[m_n_token_num ++] = m_analyzer.Get_Token();
	}
	// nate tokeny

	for(i = 0; i < m_n_token_num; i ++) {
		if(m_p_token[i].id != Ident && m_p_token[i].id != EndStmt)
			return Parser_Expression;
		// vechno mus zanat vrazem

		if(++ i == m_n_token_num)
			return Parser_Expression;
		if(m_p_token[i].id == EndStmt && i > 2)
			continue;
		// ped tm by mlo bt "<ident> = <ident>"

		if(m_p_token[i].id != Assign)
			return Parser_Expression;
		// po <ident> nslednuje "="

		if(++ i == m_n_token_num)
			return Parser_Expression;
		if(m_p_token[i].id == Ident) {
			i --;
			continue;
		}
		// pokud je po "=" dal ident, vezmeme ho jako prvn

		if(__b_IsValue(m_p_token[i].id)) {
			if(++ i == m_n_token_num)
				return Parser_Expression;
			if(m_p_token[i].id == EndStmt)
				continue;
			else
				return Parser_Expression;
		}
		// pokud je po "=" hodnota, mus za n bt stednk a potom vraz kon

		if(m_p_token[i].id == Begin) {
			if(m_p_token[++ i].id == End)
				return Parser_Expression;
			// przdn pole

			while(m_p_token[i].id != End) {
				if(!__b_IsValue(m_p_token[i].id) && m_p_token[i].id != Ident)
					return Parser_Expression;
				if(++ i == m_n_token_num)
					return Parser_Expression;
				if(m_p_token[i ++].id != Colon)
					return Parser_Expression;
				if(i == m_n_token_num)
					return Parser_Expression;
			};
			// zkontroluje, jestli je v podku obsah pole

			if(++ i == m_n_token_num)
				return Parser_Expression;
			if(m_p_token[i].id == EndStmt)
				continue;
			else
				return Parser_Expression;
			// stednk
		}
		// pokud je to pole, mus se stdat <ident> nebo hodnota s "," a po "};"

		return Parser_Expression;
	}
	// zkontroluje, jestli jdou sprvn po sob

	return Parser_Ok;
}

// uvoln vekerou pam
int CParser::Free()
{
	int i;

	if(m_p_token) {
		for(i = 0; i < m_n_token_num; i ++) {
			if(m_p_token[i].id == SValue || m_p_token[i].id == Ident)
				delete[] m_p_token[i].s_value;
		}
		// pokud je to string / ident, uvoln buffer se znaky

		delete[] m_p_token;
	}
	m_p_token = NULL;
	m_n_token_num = 0;
	//
	if(!m_analyzer.b_Stream())
		return Parser_NoFile;
	m_analyzer.Close_Stream();

	return Parser_Ok;
}

// vrt chybu syntaxe, pokud ji najde lexer
int CParser::Get_SyntaxError(char *p_error, int n_max_len)
{
	if(!m_analyzer.b_Stream())
		return Parser_NoFile;
	if(!m_analyzer.Get_LastError(p_error, n_max_len))
		return Parser_NoError;

	return Parser_Ok;
}

// te tokeny a hled "<ident> ="
// dal ten token bude hodnota
int CParser::FindKey(char *p_key_name, int &n_index)
{
	int i;

	if(!m_n_token_num)
		return Parser_NoFile;
	
	for(i = 0; i < m_n_token_num; i ++) {
		if(m_p_token[i].id == Ident && !strcmp(m_p_token[i].s_value, p_key_name)) {
			if(m_p_token[i + 1].id == Assign) {
				n_index = i + 2;
				return Parser_Ok;
			}
		}
		// pokud nael jmno
	}

	return Parser_Error;
}

// pete integerovskou hodnotu
int CParser::GetValue(char *p_key_name, int &r_n_value)
{
	static int n_level = 0;
	int n_result;
	int n_index;

	if(n_level ++ > 255) {
		n_level --;
		return Parser_CyclicRef;
	}
	// kontroluje cyklick / dlouh reference

	if((n_result = FindKey(p_key_name, n_index)) != Parser_Ok) {
		n_level --;
		return n_result;
	}
	// najde vskyt tokenu

	switch(m_p_token[n_index].id) {
	case Ident:
		n_level --;
		return GetValue(m_p_token[n_index].s_value, r_n_value);

	case CValue:
	case IValue:
		r_n_value = m_p_token[n_index].n_value;
		n_level --;
		return Parser_Ok;
	}
	// pete dal token
	
	n_level --;

	return Parser_BadType;
}

// pete char hodnotu
int CParser::GetValue(char *p_key_name, char &r_c_value)
{
	static int n_level = 0;
	int n_result;
	int n_index;

	if(n_level ++ > 255) {
		n_level --;
		return Parser_CyclicRef;
	}
	// kontroluje cyklick / dlouh reference

	if((n_result = FindKey(p_key_name, n_index)) != Parser_Ok) {
		n_level --;
		return n_result;
	}
	// najde vskyt tokenu

	switch(m_p_token[n_index].id) {
	case Ident:
		n_level --;
		return GetValue(m_p_token[n_index].s_value, r_c_value);

	case CValue:
	case IValue:
		r_c_value = (char)m_p_token[n_index].n_value;
		n_level --;
		return Parser_Ok;
	}
	// pete dal token
	
	n_level --;

	return Parser_BadType;
}

// pete float hodnotu
int CParser::GetValue(char *p_key_name, float &r_f_value)
{
	static int n_level = 0;
	int n_result;
	int n_index;

	if(n_level ++ > 255) {
		n_level --;
		return Parser_CyclicRef;
	}
	// kontroluje cyklick / dlouh reference

	if((n_result = FindKey(p_key_name, n_index)) != Parser_Ok) {
		n_level --;
		return n_result;
	}
	// najde vskyt tokenu

	switch(m_p_token[n_index].id) {
	case Ident:
		n_level --;
		return GetValue(m_p_token[n_index].s_value, r_f_value);

	case FValue:
		r_f_value = m_p_token[n_index].f_value;
		n_level --;
		return Parser_Ok;
	case IValue:
		r_f_value = (float)m_p_token[n_index].n_value;
		n_level --;
		return Parser_Ok;
	}
	// pete dal token
	
	n_level --;

	return Parser_BadType;
}

// pete string
int CParser::GetValue(char *p_key_name, char *p_str_value, int n_max_size)
{
	static int n_level = 0;
	int n_result;
	int n_index;

	if(n_level ++ > 255) {
		n_level --;
		return Parser_CyclicRef;
	}
	// kontroluje cyklick / dlouh reference

	if((n_result = FindKey(p_key_name, n_index)) != Parser_Ok) {
		n_level --;
		return n_result;
	}
	// najde vskyt tokenu

	switch(m_p_token[n_index].id) {
	case Ident:
		n_level --;
		return GetValue(m_p_token[n_index].s_value, p_str_value, n_max_size);

	case SValue:
		strncpy(p_str_value, m_p_token[n_index].s_value, n_max_size);
		n_level --;
		return Parser_Ok;
	}
	// pete dal token
	
	n_level --;

	return Parser_BadType;
}

int CParser::GetStrLength(char *p_key_name, int &n_length)
{
	static int n_level = 0;
	int n_result;
	int n_index;

	if(n_level ++ > 255) {
		n_level --;
		return Parser_CyclicRef;
	}
	// kontroluje cyklick / dlouh reference

	if((n_result = FindKey(p_key_name, n_index)) != Parser_Ok) {
		n_level --;
		return n_result;
	}
	// najde vskyt tokenu

	switch(m_p_token[n_index].id) {
	case Ident:
		n_level --;
		return GetStrLength(m_p_token[n_index].s_value, n_length);

	case SValue:
		n_length = strlen(m_p_token[n_index].s_value);
		n_level --;
		return Parser_Ok;
	}
	// pete dal token
	
	n_level --;

	return Parser_BadType;
}