/*
								+----------------------------------+
								|                                  |
								|       ***  Text parser  ***      |
								|                                  |
								|   Copyright  -tHE SWINe- 2003   |
								|                                  |
								|             Parser.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __TXT_PARSER_INCLUDED
#define __TXT_PARSER_INCLUDED

enum {
	Parser_Ok,			// ok
	Parser_NoError,		// no error (when calling Get_SyntaxError())
	Parser_Error,		// generic error
	Parser_NoFile,		// no file
	Parser_Syntax,		// syntax error
	Parser_Expression,	// expression syntax error
	Parser_Memory,		// not enough memory
	Parser_BadType,		// bad type
	Parser_CyclicRef	// cyclic reference
};

enum Token_Id {
	Begin, End, EndStmt, Ident,
	IValue, CValue, FValue,
	SValue, Assign, Colon, _Unknown_
};
// token names

struct Token {
	Token_Id id;
	//
	int n_line;
	//
	float f_value;
	long n_value;
	char *s_value;
};
// token

class CLexicalAnalyzer {
	int m_n_stream_len;
	int m_n_stream_pos;
	char *m_p_stream;
	//
	char m_s_last_error[256];

public:
	CLexicalAnalyzer();
	virtual ~CLexicalAnalyzer();
	//
	int Open_Stream(FILE *p_fr);
	int Open_Stream(char *p_stream_str, int n_stream_len);
	void Reset_Stream();
	int Close_Stream();
	//
	int b_End_of_Stream();
	int b_Stream();
	//
	Token Get_Token();
	//
	int Get_LastError(char *p_buffer, int n_buffer_length);

private:
	int Cur_Line();
	int GetChar();
	void PushChar();
	//
	void Error(char *p_str, int n_line);
};

class CParser {
	CLexicalAnalyzer m_analyzer;
	//
	int m_n_token_num;
	Token *m_p_token;

public:
	CParser();
	~CParser();
	//
	int Parse(FILE *p_file);
	int Parse(char *p_stream, int n_stream_len);
	//
	int Free();
	//
	int Get_SyntaxError(char *p_error, int n_max_len);
	//
	int GetValue(char *p_key_name, int &r_n_value);
	int GetValue(char *p_key_name, char &r_c_value);
	int GetValue(char *p_key_name, float &r_f_value);
	int GetValue(char *p_key_name, char *p_str_value, int n_max_size);
	int GetStrLength(char *p_key_name, int &n_length);

private:
	int Get_Tokens();
	//
	int FindKey(char *p_key_name, int &n_index);
};

#endif